#include "CityGrid.h"
#include "CitySettings.h"

// Sets default values
ACityGrid::ACityGrid()
{
 	// Set this actor to call Tick() every frame.  You can turn this off to improve performance if you don't need it.
	PrimaryActorTick.bCanEverTick = true;

	// Generate base geometry	
	mesh = CreateDefaultSubobject<UProceduralMeshComponent>(TEXT("CityMesh"));
	static ConstructorHelpers::FObjectFinder<UMaterial> MaterialOb(TEXT("Material'/Game/VertexMaterial.VertexMaterial'"));
	material = MaterialOb.Object;
	// All transformations are applied to this mesh
	RootComponent = mesh;

	//generate city
	layout = nullptr;
	generateMesh = false;
	speed = 2;
	height = 20;
	spacing = 20;
	displayCustomersInRow = 3;
	limitCustomersInside = 9;

	//individual
	personalSpaceRangeMin = 1.0;
	personalSpaceRangeMax = 5.0;
	pushFactorRangeMin = 0.0;
	pushFactorRangeMax = 2.0;
	foodHungerDeltaMin = 0.1;
	foodHungerDeltaMax = 0.1;
	foodEatingDeltaMin = 1.0;
	foodEatingDeltaMax = 1.0;
	foodCapacityMin = 100.0;
	foodCapacityMax = 100.0;
	patienceLimitMin = 1000.0;
	patienceLimitMax = 1000.0;
	patienceDeltaMin = 0.5;
	patienceDeltaMax = 0.5;

	//common
	perceptionRadius = 15;
	walkCheckpointTolerance = 4;
	agentRadius = 1;
	talkingDistance = 4;
	weightWallRepulsion = 1.0;
	weightAgentRepulsion = 1.0;
	weightWallAvoidance = 1.0;
	weightAgentAvoidance = 1.0;
	weightSteeringForce = 10.0;
	hungryLevel = 5;

	//initiate agents mesh
	agentMesh = CreateDefaultSubobject<UInstancedStaticMeshComponent>(TEXT("AgentMesh")); // A component that efficiently renders multiple instances of the same StaticMesh.
	agentMesh->NumCustomDataFloats = 1; // Defines the number of floats that will be available per instance for custom data
	agentMesh->RegisterComponent();

	static ConstructorHelpers::FObjectFinder<UMaterial> MaterialAgentOb(TEXT("Material'/Game/AgentMaterial.AgentMaterial'"));
	UStaticMesh* agentMeshInstance = ConstructorHelpers::FObjectFinder<UStaticMesh>(TEXT("StaticMesh'/Engine/BasicShapes/Sphere.Sphere'")).Object;

	agentMeshInstance->SetMaterial(0, MaterialAgentOb.Object);

	agentMesh->SetStaticMesh(agentMeshInstance);

	map = nullptr;
}

// Called when the game starts or when spawned
void ACityGrid::BeginPlay()
{
	Super::BeginPlay();

	initCity();
	initAgents();
}

// Called every frame
void ACityGrid::Tick(float DeltaTime)
{
	Super::Tick(DeltaTime);

	std::vector<Agent*> others;
	
	//update agents, compute forces
	for (size_t i = 0; i < agents.size(); ++i)
	{
		map->getOccupantsForPosition(others, agents[i]->getPosition(), perceptionRadius);
		agents[i]->forcesCompute(map, others);
	}
	
	//reset occupants
	map->emptyOccupants();
	

	FVector loc;
	Tiles* tile;

	//update agent geometry position
	for (size_t i = 0; i < agents.size(); ++i)
	{
		agents[i]->forcesApply(DeltaTime * speed * 10.0f);
		agentMesh->UpdateInstanceTransform(i, agents[i]->transform, true, i == (agents.size() - 1));
		agentMesh->SetCustomDataValue(i, 0, agents[i]->getFeatureState(), i == (agents.size() - 1));

		loc = agents[i]->getPosition();
		tile = map->getTileForLocation(loc);

		//to prevent all evil
		if (tile)
			tile->addOccupant(agents[i]);
		else
		{
			agents.erase(agents.begin() + i);
			i--;
		}
	}
}

void clampMinMax(float& min, float& max)
{
	min = std::min(min, max);
	max = std::max(min, max);
}

void ACityGrid::OnConstruction(const FTransform& Transform)
{
	if (generateMesh && layout)
		initCity();

	//clamp parameters
	clampMinMax(personalSpaceRangeMin, personalSpaceRangeMax);
	clampMinMax(pushFactorRangeMin, pushFactorRangeMax);
	clampMinMax(foodHungerDeltaMin, foodHungerDeltaMax);
	clampMinMax(foodEatingDeltaMin, foodEatingDeltaMax);
	clampMinMax(foodCapacityMin, foodCapacityMax);
	clampMinMax(patienceLimitMin, patienceLimitMax);
	clampMinMax(patienceDeltaMin, patienceDeltaMax);
}

void ACityGrid::initCity()
{
	generateMesh = false;
	if (map)
		delete map;

	CitySettings settings;
	settings.spacing = spacing;
	settings.height = height;
	settings.material = material;
	settings.layout = layout;
	settings.mesh = mesh;
	settings.foodmarketAgentsRow = displayCustomersInRow;
	settings.foodmarketLimit = limitCustomersInside;

	// create CityMap
	map = new CityMap();
	map->generate(settings);
}

void ACityGrid::initAgents()
{
	// common parameters for agents
	AgentSettings settings; 
	settings.size = spacing;
	settings.radius = agentRadius;
	settings.hungryLevel = hungryLevel;
	settings.walkCheckpointTolerance = walkCheckpointTolerance;
	settings.talkingDistance = talkingDistance;
	settings.weightWallRepulsion = weightWallRepulsion;
	settings.weightAgentRepulsion = weightAgentRepulsion;
	settings.weightWallAvoidance = weightWallAvoidance;
	settings.weightAgentAvoidance = weightAgentAvoidance;
	settings.weightSteeringForce = weightSteeringForce;

	// random generator helpers for individual parameters
	std::default_random_engine generator;
	std::uniform_real_distribution<float> personalSpace(personalSpaceRangeMin, personalSpaceRangeMax);
	std::uniform_real_distribution<float> push(pushFactorRangeMin, pushFactorRangeMax);
	std::uniform_real_distribution<float> limit(foodCapacityMin, foodCapacityMax);
	std::uniform_real_distribution<float> hunger(foodHungerDeltaMin, foodHungerDeltaMax);
	std::uniform_real_distribution<float> eat(foodEatingDeltaMin, foodEatingDeltaMax);
	std::uniform_real_distribution<float> patienceLimit(patienceLimitMin, patienceLimitMax);
	std::uniform_real_distribution<float> patienceDelta(patienceDeltaMin, patienceDeltaMax);

	// individual parameters for agents
	const std::vector<Tiles*>& buildings = map->getBuildings();

	Agent* agent;
	for (size_t i = 0; i < buildings.size(); ++i)
	{
		settings.personalSpace = personalSpace(generator);
		settings.pushing = push(generator);
		settings.foodCapacity = limit(generator);
		settings.foodHungerDelta = hunger(generator);
		settings.foodEatingDelta = eat(generator);
		settings.patienceLimit = patienceLimit(generator);
		settings.patienceDelta = patienceDelta(generator);

		agent = new Agent((BuildingTile*)buildings[i], GetTransform().GetLocation(), settings); // Agent::Agent(BuildingTile* home, FVector woffset, AgentSettings& settings)
		agentMesh->AddInstance(agent->transform);
		agents.push_back(agent);
	}
}


